from pybuoy.api.forecasts import Forecasts
from pybuoy.api.realtime import Realtime
from pybuoy.api.stations import Stations


class Buoy:
    """Buoy class provides convenient access to NDBC's API.

    Instances of this class are the gateway to interacting with National
    Data Buoy Center (NDBC) through `pybuoy`. The canonical way to get
    an instance of this class is via:

    .. code-block:: python

        from pybuoy import Buoy

        buoy = Buoy() # creates instance to access data from NDBC

        meteorological_data = buoy.realtime.get(station_id="44065")

        all_active_NDBC_stations = buoy.stations.get_active()
    """

    def __init__(self):
        self._forecasts = Forecasts()
        self._realtime = Realtime()
        self._stations = Stations()

    @property
    def forecasts(self) -> Forecasts:
        return self._forecasts

    @property
    def realtime(self) -> Realtime:
        return self._realtime

    @property
    def stations(self) -> Stations:
        return self._stations



from typing import Literal, overload

from pybuoy.api.base import ApiBase
from pybuoy.const import API_PATH, Endpoints, RealtimeDatasets, RealtimeDatasetsValues
from pybuoy.observation.observations import (
    MeteorologicalObservations,
    WaveSummaryObservations,
)


class Realtime(ApiBase):
    @overload
    def get(
        self,
        station_id: str,
        dataset: Literal["spec"],
    ) -> WaveSummaryObservations: ...

    @overload
    def get(
        self, station_id: str, dataset: Literal["txt"] = "txt"
    ) -> MeteorologicalObservations: ...

    # TODO: consider mapping str literal to dataset
    def get(
        self,
        station_id: str,
        dataset: RealtimeDatasetsValues = RealtimeDatasets.txt.value,
    ):
        """Get realtime data from the NDBC.

        There are nine different data sources:
            - data_spec     Raw Spectral Wave Data
            - ocean         Oceanographic Data
            - spec          Spectral Wave Summary Data
            - supl          Supplemental Measurements Data
            - swdir         Spectral Wave Data (alpha1)
            - swdir2        Spectral Wave Data (alpha2)
            - swr1          Spectral Wave Data (r1)
            - swr2          Spectral Wave Data (r2)
            - txt           Standard Meteorological Data

        Args:
            buoy_id (str): id of buoy
            dataset (str): one of the aforementioned data sources
        """
        if not station_id:
            raise ValueError("station_id must be provided")

        dataset_options = {
            "data_spec",
            "ocean",
            "spec",
            "supl",
            "swdir",
            "swdir2",
            "swr1",
            "swr2",
            "txt",
        }

        if dataset not in dataset_options:
            raise ValueError(f"Dataset must be one of {', '.join(dataset_options)}")

        url = f"{API_PATH[Endpoints.REALTIME.value]}/{station_id}.{dataset}"

        return self.parse(data=self.make_request(url), dataset=dataset)



from collections import defaultdict
from datetime import datetime as dt
from json import loads
from typing import Optional

import lxml.etree as ET  # type: ignore

from pybuoy.api.base import ApiBase
from pybuoy.const import API_PATH, Endpoints
from pybuoy.observation import ForecastObservation, ForecastObservations
from pybuoy.unit_mappings import NO_NUMERIC_VALUE


class Forecasts(ApiBase):
    # https://graphical.weather.gov/xml/rest.php
    def get(self, lat: float, lon: float, start_date: str, end_date: str):
        dwml_data = self.forecast_by_lat_lon(
            lat=lat,
            lon=lon,
            start_date=dt.fromisoformat(start_date),
            end_date=dt.fromisoformat(end_date),
        )
        grouped = self.__group_parameters_by_time(dwml_data["data"]["parameters"])

        return ForecastObservations(
            observations=[
                ForecastObservation(grouped[key], dt.fromisoformat(key))
                for key in grouped.keys()
            ],
            repr_limit=10,
        )

    def forecast_by_lat_lon(
        self,
        lat,
        lon,
        start_date: Optional[dt] = None,
        end_date: Optional[dt] = None,
        metric=False,
    ):
        return self._daily_forecast_from_location_info(
            location_info=[lat, lon],
            start_date=start_date,
            end_date=end_date,
            metric=metric,
        )

    def _daily_forecast_from_location_info(
        self,
        location_info,
        start_date: Optional[dt] = None,
        end_date: Optional[dt] = None,
        metric=False,
    ):
        if not start_date:
            start_date = dt.today()

        lat, lon = location_info
        params = {
            "whichClient": "NDFDgen",
            "lat": lat,
            "lon": lon,
            "product": "time-series",
            "begin": start_date.isoformat(),
            "end": None if end_date is None else end_date.isoformat(),
            "Unit": "m" if metric else "e",
            "wspd": "wspd",
            "wdir": "wdir",
            "waveh": "waveh",
            "wgust": "wgust",
            "Submit": "Submit",
        }
        response = self.make_request(API_PATH[Endpoints.FORECASTS.value], params=params)
        xml_root = ET.fromstring(response.encode("utf-8"))
        parsed = self.xslt_transform(xml_root)

        return loads(str(parsed))

    def __group_parameters_by_time(self, parameters: dict):
        """Group DWML as XSLT by observation recorded.

        Args:
            parameters (dict): XSLT object

        Returns:
            (
                str,
                defaultdict(
                    ForecastKey,
                    {
                        "value": int,
                        "label": str,
                        "units": str,
                    }
                )
            )

        """
        # TODO: refactor
        # XSLT flatten data instead?
        ele_map = {
            "direction": ("wind-direction",),
            "wind": (
                "gust",
                "sustained",
            ),
            "water-state": ("significant",),
        }
        default_val = {"value": NO_NUMERIC_VALUE}
        groupedby_timestamp: dict[str, dict] = defaultdict(
            lambda: defaultdict(
                dict,
                wave_height=default_val,
                wind_direction=default_val,
                wind_speed_gust=default_val,
                wind_speed=default_val,
            )
        )
        for k, v in parameters.items():
            for key in v.keys():
                if key in ele_map[k]:
                    for period in parameters[k][key]["periods"]:
                        timestamp = period["start-time"]
                        new_key = "_".join(
                            parameters[k][key]["name"].lower().split(" ")
                        )
                        groupedby_timestamp[timestamp][new_key] = {
                            "value": period["value"],
                            # TODO: transform to proper label here
                            "label": parameters[k][key]["name"],
                            "units": parameters[k][key]["units"],
                        }

        return groupedby_timestamp



from pybuoy.observation.observation import MeteorologicalObservation  # noqa: F401
from pybuoy.observation.observation import (  # noqa: F401
    ForecastObservation,
    WaveSummaryObservation,
)
from pybuoy.observation.observation_datum import (  # noqa: F401
    ObservationFloatDatum,
    ObservationStringDatum,
)
from pybuoy.observation.observations import (  # noqa: F401
    ForecastObservations,
    MeteorologicalObservations,
    WaveSummaryObservations,
)



"""List of API endpoints pybuoy knows about."""

_BASE_URL = "https://www.ndbc.noaa.gov"
_FORECASTS_URL = "https://graphical.weather.gov/xml/sample_products/browser_interface/ndfdXMLclient.php"  # noqa: E501

API_PATH: dict[str, str] = {
    "active_stations": f"{_BASE_URL}/activestations",
    "forecasts": f"{_FORECASTS_URL}",
    "realtime2": f"{_BASE_URL}/data/realtime2",
}



from datetime import datetime
from typing import Optional

from pybuoy.observation.observation_datum import (
    ObservationFloatDatum,
    ObservationStringDatum,
)
from pybuoy.unit_mappings import (
    FORECAST,
    METEOROLOGICAL,
    WAVE_SUMMARY,
    MeteorologicalKey,
    WaveSummaryKey,
)

# TODO: support dynamic typing and add slots for efficiency


class BaseObservation:
    """BaseObservation class for `Buoy` readings by datetime."""

    def __init__(self, datetime: Optional[datetime] = None):
        """Initialize BaseObservation record with datetime.

        Args:
            datetime (datetime): UTC time of value. Defaults to None.
        """
        self._datetime = datetime

    @property
    def datetime(self) -> datetime | None:
        """Return when this observation was made."""
        return self._datetime

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ", ".join("{}={!r}".format(k, v) for k, v in self.__dict__.items()),
        )

    def __str__(self):
        return f"Observation({self.datetime})"


class MeteorologicalObservation(BaseObservation):
    """Encapsulates `Buoy` meteorological data."""

    def __init__(
        self,
        values: dict[MeteorologicalKey, str],
        datetime: Optional[datetime] = None,
    ):
        """Initialize Observation record with relevant metadata.

        Args:
            values (dict[str,str]): recorded weather data.
            datetime (datetime): UTC time of value. Defaults to None.
        """
        super().__init__(datetime=datetime)

        for key in METEOROLOGICAL:
            new_key = "_".join(METEOROLOGICAL[key]["label"].lower().split(" "))
            setattr(
                self,
                f"_{new_key}",
                ObservationStringDatum(values[key], METEOROLOGICAL[key]),
            ) if METEOROLOGICAL[key]["unit"] == "string" else setattr(
                self,
                f"_{new_key}",
                ObservationFloatDatum(values[key], METEOROLOGICAL[key]),
            )

    @property
    def wind_direction(self) -> ObservationFloatDatum:
        """Return observed wind direction."""
        return self._wind_direction  # type: ignore # this is dynamically set

    @property
    def wind_speed(self) -> ObservationFloatDatum:
        """Return observed wind speed."""
        return self._wind_speed  # type: ignore # this is dynamically set

    @property
    def wind_gust(self) -> ObservationFloatDatum:
        """Return observed wind gust."""
        return self._wind_gust  # type: ignore # this is dynamically set

    @property
    def wave_height(self) -> ObservationFloatDatum:
        """Return observed wave height."""
        return self._wave_height  # type: ignore # this is dynamically set

    @property
    def dominant_wave_period(self) -> ObservationFloatDatum:
        """Return observed dominant wave period."""
        return self._dominant_wave_period  # type: ignore # this is dynamically set

    @property
    def average_wave_period(self) -> ObservationFloatDatum:
        """Return observed average wave period."""
        return self._average_wave_period  # type: ignore # this is dynamically set

    @property
    def wave_direction(self) -> ObservationFloatDatum:
        """Return observed wave direction."""
        return self._wave_direction  # type: ignore # this is dynamically set

    @property
    def sea_level_pressure(self) -> ObservationFloatDatum:
        """Return observed sea level pressure."""
        return self._sea_level_pressure  # type: ignore # this is dynamically set

    @property
    def air_temperature(self) -> ObservationFloatDatum:
        """Return observed air temperature."""
        return self._air_temperature  # type: ignore # this is dynamically set

    @property
    def water_temperature(self) -> ObservationFloatDatum:
        """Return observed water temperature."""
        return self._water_temperature  # type: ignore # this is dynamically set

    @property
    def dewpoint_temperature(self) -> ObservationFloatDatum:
        """Return observed dewpoint temperature."""
        return self._dewpoint_temperature  # type: ignore # this is dynamically set

    @property
    def visibility(self) -> ObservationFloatDatum:
        """Return observed visibility."""
        return self._visibility  # type: ignore # this is dynamically set

    @property
    def pressure_tendency(self) -> ObservationFloatDatum:
        """Return observed pressure tendency."""
        return self._pressure_tendency  # type: ignore # this is dynamically set

    @property
    def tide(self) -> ObservationFloatDatum:
        """Return observed tide."""
        return self._tide  # type: ignore # this is dynamically set


class WaveSummaryObservation(BaseObservation):
    """Encapsulates `Buoy` wave summary data."""

    def __init__(
        self,
        values: dict[WaveSummaryKey, str],
        datetime: Optional[datetime] = None,
    ):
        """Initialize Observation record with relevant metadata.

        Args:
            values (dict[str,str]): recorded weather data.
            datetime (datetime): UTC time of value. Defaults to None.
        """
        super().__init__(datetime=datetime)

        for key in WAVE_SUMMARY:
            new_key = "_".join(WAVE_SUMMARY[key]["label"].lower().split(" "))
            setattr(
                self,
                f"_{new_key}",
                ObservationStringDatum(values[key], WAVE_SUMMARY[key]),
            ) if WAVE_SUMMARY[key]["unit"] == "string" else setattr(
                self,
                f"_{new_key}",
                ObservationFloatDatum(values[key], WAVE_SUMMARY[key]),
            )

    @property
    def significant_wave_height(self) -> ObservationFloatDatum:
        """Return observed wave height."""
        return self._significant_wave_height  # type: ignore # this is dynamically set

    @property
    def swell_height(self) -> ObservationFloatDatum:
        """Return observed swell height."""
        return self._swell_height  # type: ignore # this is dynamically set

    @property
    def swell_period(self) -> ObservationFloatDatum:
        """Return observed swell period."""
        return self._swell_period  # type: ignore # this is dynamically set

    @property
    def wind_wave_height(self) -> ObservationFloatDatum:
        """Return observed wind wave height."""
        return self._wind_wave_height  # type: ignore # this is dynamically set

    @property
    def wind_wave_period(self) -> ObservationFloatDatum:
        """Return observed wind wave period."""
        return self._wind_wave_period  # type: ignore # this is dynamically set

    @property
    def swell_direction(self) -> ObservationStringDatum:
        """Return observed swell direction."""
        return self._swell_direction  # type: ignore # this is dynamically set

    @property
    def wind_wave_direction(self) -> ObservationStringDatum:
        """Return observed wind wave direction."""
        return self._wind_wave_direction  # type: ignore # this is dynamically set

    @property
    def steepness(self) -> ObservationStringDatum:
        """Return observed wave steepness."""
        return self._steepness  # type: ignore # this is dynamically set

    @property
    def average_wave_period(self) -> ObservationFloatDatum:
        """Return observed average wave period."""
        return self._average_wave_period  # type: ignore # this is dynamically set

    @property
    def dominant_wave_direction(self) -> ObservationFloatDatum:
        """Return observed direction of waves at dominant period."""
        return self._dominant_wave_direction  # type: ignore # this is dynamically set


class ForecastObservation(BaseObservation):
    """Encapsulates forecasted `Buoy` data."""

    def __init__(
        self,
        # TODO: fix dict key type
        values: dict[str, dict],
        datetime: Optional[datetime] = None,
    ):
        super().__init__(datetime=datetime)
        for key in FORECAST:
            new_key = "_".join(FORECAST[key]["label"].lower().split(" "))
            default_val = {"value": "nan"}  # TODO: refactor
            value = values.get(key.name, default_val)
            value_value = value.get("value", "nan")  # TODO: rename
            setattr(
                self,
                f"_{new_key}",
                ObservationFloatDatum(value_value, FORECAST[key]),
            )

    def __str__(self):
        return f"Prediction({self.datetime})"

    @property
    def wind_direction(self) -> ObservationFloatDatum:
        """Return observed wind direction."""
        return self._wind_direction  # type: ignore # this is dynamically set

    @property
    def wind_speed(self) -> ObservationFloatDatum:
        """Return observed wind speed."""
        return self._wind_speed  # type: ignore # this is dynamically set

    @property
    def wind_gust(self) -> ObservationFloatDatum:
        """Return observed wind gust."""
        return self._wind_gust  # type: ignore # this is dynamically set

    @property
    def wave_height(self) -> ObservationFloatDatum:
        """Return observed wave height."""
        return self._wave_height  # type: ignore # this is dynamically set



"""Provide the ParserMixin class."""

from collections import defaultdict
from datetime import datetime as dt
from os.path import dirname, join
from typing import Any
from xml.etree.ElementTree import Element, fromstring

import lxml.etree as ET  # type: ignore

from pybuoy.const import RealtimeDatasets, RealtimeDatasetsValues
from pybuoy.exceptions import NDBCException
from pybuoy.observation import MeteorologicalObservation, WaveSummaryObservation
from pybuoy.observation.observations import (
    MeteorologicalObservations,
    WaveSummaryObservations,
)
from pybuoy.unit_mappings import MeteorologicalKey, WaveSummaryKey


class ParserMixin:
    """Parser mixin supports handling of third-party data."""

    def __init__(self) -> None:
        self.__set_xslt_transform()

    def etree_to_dict(self, element: Element):
        """Parse XML Element to dictionary."""
        from_etree: dict[str, dict[str, Any] | Any] = {
            element.tag: {} if element.attrib else None
        }
        children = list(element)
        if children:
            parsed_children = defaultdict(list)
            for child in map(self.etree_to_dict, children):
                for key, value in child.items():
                    parsed_children[key].append(value)
            from_etree = {
                element.tag: {
                    key: value[0] if len(value) == 1 else value
                    for key, value in parsed_children.items()
                }
            }
        if element.attrib:
            from_etree[element.tag].update(
                (key, value) for key, value in element.attrib.items()
            )
        if element.text:
            text = element.text.strip()
            if children or element.attrib:
                if text:
                    from_etree[element.tag]["text"] = text
            else:
                from_etree[element.tag] = text
        return from_etree

    def parse(
        self,
        data: str,
        dataset: RealtimeDatasetsValues,
    ) -> MeteorologicalObservations | WaveSummaryObservations | str:
        obs = self.__clean_realtime_data(data=data, dataset=dataset)
        # TODO: refine error handling
        if obs is None or len(obs) == 0:
            raise NDBCException

        match dataset:
            case RealtimeDatasets.spec.value:
                # ? overloading can be an alternative
                wave_obs: list[WaveSummaryObservation] = obs  # type: ignore
                return WaveSummaryObservations(observations=wave_obs)
            case RealtimeDatasets.txt.value:
                met_obs: list[MeteorologicalObservation] = obs  # type: ignore
                return MeteorologicalObservations(observations=met_obs)
            case _:
                return data

    @property
    def xslt_transform(self) -> ET.XSLT:
        return self._xslt_transform

    # TODO: consider station class/dto like observations
    def _clean_activestation_data(self, data: str):
        xml_tree_root = fromstring(data.encode("utf-8"))
        # TODO: consider incorporating `etree_to_dict`
        return [dict(el.items()) for el in xml_tree_root.findall("station")]

    # TODO: fix missing latest row
    def __clean_realtime_data(
        self, data: str, dataset: RealtimeDatasetsValues
    ) -> list[MeteorologicalObservation] | list[WaveSummaryObservation]:
        if data is None:
            # TODO: handle when request not successful
            raise NDBCException

        # TODO: consider csv module
        rows = data.strip().split("\n")
        headers = [" ".join(row.split()).split(" ") for row in rows[0:2]]
        header_offset = 5  # end of datetime columns
        headers_without_dates = headers[0][header_offset:]

        obs = []
        for row in rows[2:]:
            record_array = " ".join(row.split()).split(" ")
            date_recorded = dt(
                int(record_array[0]),
                int(record_array[1]),
                int(record_array[2]),
                int(record_array[3]),
                int(record_array[4]),
            )
            # TODO: investigate adding Strategy Pattern here
            if dataset == RealtimeDatasets.txt.value:
                observation = self.__parse_meteorological_record(
                    date_recorded=date_recorded,
                    headers=headers_without_dates,
                    records=record_array[header_offset:],
                )
            elif dataset == RealtimeDatasets.spec.value:
                observation = self.__parse_wave_summary_record(
                    date_recorded=date_recorded,
                    headers=headers_without_dates,
                    records=record_array[header_offset:],
                )
            obs.append(observation) if observation is not None else None

        return obs

    def __parse_meteorological_record(self, date_recorded: dt, headers, records):
        values: dict[MeteorologicalKey, str] = {
            MeteorologicalKey[headers[idx]]: value for idx, value in enumerate(records)
        }
        return MeteorologicalObservation(values, date_recorded)

    # TODO: refactor (e.g., combine with __parse_meteorological_record)
    def __parse_wave_summary_record(self, date_recorded: dt, headers, records):
        values: dict[WaveSummaryKey, str] = {
            WaveSummaryKey[headers[idx]]: value for idx, value in enumerate(records)
        }
        return WaveSummaryObservation(values, date_recorded)

    def __set_xslt_transform(self, xsl_filename="noaa.xsl"):
        forecast_xslt = ET.parse(join(dirname(__file__), xsl_filename))
        self._xslt_transform = ET.XSLT(forecast_xslt)



"""pybuoy exception classes."""

from requests import HTTPError  # type: ignore


class BaseException(Exception):
    """Base Exception that all other exception classes extend."""

    ...


class HTTPException(HTTPError): ...


class NDBCException(BaseException): ...


class NOAAException(BaseException): ...


class DWMLException(NOAAException):
    """NOAA DWML Exception.

    ref: https://graphical.weather.gov/xml/DWMLgen/schema/DWML.xsd
    """

    ...



"""pybuoy constants."""

from enum import Enum
from typing import Literal

from pybuoy.endpoints import API_PATH as API_PATH  # noqa: F401


class Endpoints(Enum):
    ACTIVE_STATIONS = "active_stations"
    FORECASTS = "forecasts"
    REALTIME = "realtime2"


class RealtimeDatasets(Enum):
    data_spec = "data_spec"
    ocean = "ocean"
    spec = "spec"
    supl = "supl"
    swdir = "swdir"
    swdir2 = "swdir2"
    swr1 = "swr1"
    swr2 = "swr2"
    txt = "txt"


RealtimeDatasetsValues = Literal[
    "data_spec",
    "ocean",
    "spec",
    "supl",
    "swdir",
    "swdir2",
    "swr1",
    "swr2",
    "txt",
]

__version__ = "0.5.6"
